/*
 * File: modedet.c
 * Detection of the modes of a distribution and extraction of the capacity mode.
 * 
 * Copyright 2003 Mathieu Goutelle (INRIA, France)
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License, version 2, as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * A copy of the GNU GPL is available as /usr/doc/copyright/GPL on Debian
 * systems, or on the World Wide Web at http://www.gnu.org/copyleft/gpl.html You
 * can also obtain it by writing to the Free Software Foundation, Inc., 59
 * Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * CVSid $Id: modedet.c,v 1.1 2003/12/15 13:59:23 mgoutell Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "modedet.h"

/*
 * Detection of the modes of the measures distribution. 'step' is the width of
 * the bins used in the distribution. The modes are stored in... 'modes' and the
 * return value is the number of detected modes.
 */
int detect_modes(double *mesures, int nb_mes, double step, struct mode* modes) {
  int i, nb_bin, cpt_bin, nb_mode, thresh = 2;
  int ltmin, max, rtmin, pop;
  double avg;
  int* distrib_mes;
  double max_mes = 0;
  
  /* Looking for the biggest value */
  for (i = 0; i < nb_mes; i++)
    if (max_mes < mesures[i]) max_mes = mesures[i];

  /* Determining the measurements distribution */
  nb_bin = (int)floor(max_mes/step) + 1;
  distrib_mes = calloc(nb_bin, sizeof(int));
  for (i = 0; i < nb_mes; i++)
    distrib_mes[(int)floor(mesures[i]/step)]++;
  /*   for (i = 0; i < nb_bin; i++) */
  /*     printf("%7.2f %d\n", ((double)i+0.5)*step, distrib_mes[i]); */

  ltmin = max = rtmin = pop = 0; avg = 0.0;
  nb_mode = 0; cpt_bin = 0;

  for (i = 0; i < nb_bin; i++) {
    switch (cpt_bin) {
    case 0: /* Looking for the first value greater than the threshold */
      if (distrib_mes[i] >= thresh) {
        max = ltmin = i;
        pop = distrib_mes[i];
        avg = ((double)i+0.5)*distrib_mes[i];
        cpt_bin = 1;
      }
      break;
    case 1: /* Looking for the local maximum */
      if (distrib_mes[i] >= distrib_mes[max]-thresh && distrib_mes[i] >= thresh) {
        max = (distrib_mes[i] >= distrib_mes[max]?i:max);
        rtmin = i;
        pop += distrib_mes[i];
        avg += ((double)i+0.5)*distrib_mes[i];	
      }
      else { /* Last bin was local max*/
        if (distrib_mes[i] >= thresh) {
          rtmin = i;
          pop += distrib_mes[i];
          avg += ((double)i+0.5)*distrib_mes[i];	
          cpt_bin = 2;
        }
        else { /* End of mode */
          rtmin = (rtmin>max?rtmin:max);
          cpt_bin = 3;
        }
      }
      break;
    case 2: /* Looking for the value lower than max in a decreasive trend */
      if (distrib_mes[i]-thresh <= distrib_mes[rtmin] && distrib_mes[i] >= thresh) {
        rtmin = i;
        pop += distrib_mes[i];
        avg += ((double)i+0.5)*distrib_mes[i];	
        break;
      }
      else /* End of mode */
        cpt_bin = 3;
    case 3: /* All the parameters of the current mode are known. Store it and continue */
      modes[nb_mode].ltmin = ltmin;
      modes[nb_mode].max = max;
      modes[nb_mode].rtmin = rtmin;
      modes[nb_mode].pop = pop;
      modes[nb_mode].avg = avg/(double)pop;

      nb_mode++;
      if (nb_mode >= NB_MODES) return NB_MODES; /* Max nb of nodes reached. INCREASE IT !!! */
      if (distrib_mes[i] >= thresh) { /* Handle two consecutive modes */
        max = ltmin = i;
        pop = distrib_mes[i];
        avg = ((double)i+0.5)*distrib_mes[i];       
        cpt_bin = 1;
      }
      else
        cpt_bin = 0;
      break;
    }
    //printf("%d %d %d %d %d %d\n", i, cpt_bin, distrib_mes[i], ltmin, max, rtmin);
  }

  if (cpt_bin) { /* Terminate last mode in case of */
    if (cpt_bin == 1)
      rtmin = max;

    modes[nb_mode].ltmin = ltmin;
    modes[nb_mode].max = max;
    modes[nb_mode].rtmin = rtmin;
    modes[nb_mode].pop = pop;
    modes[nb_mode].avg = avg/(double)pop;
    nb_mode++;
  }

  return nb_mode;
}

/*
 * Extract the capacity mode from the measurements stored in the 'modes'
 * array. The previous detected CM are in the 'cm' array. For extracting CM, we
 * do need three values : the new mode, the max mode and the previous detected
 * mode. When these three values are determined, some tests are performed to try
 * to evaluate properly the capacity mode among them. You can see that with the
 * successive tests, the situation (noise, concurrent traffic and longer path) is
 * getting worst.
 */
int extract_capa_mode(struct mode *cm, int nb_cm, struct mode *modes, int nb_modes, int nb_mes) {
  int i, new_mode = nb_modes-1, max_mode = 0, last_mode = nb_modes-1, 
    end_of_noise = -1, begin_of_noise = -1;

  /* Looking for a new mode strictly below the last detected mode */ 
  while (new_mode >= 0) {
    if ((modes[new_mode].pop <= nb_mes/100) || //(modes[j].rtmin - modes[j].ltmin > 3) || 
        ((nb_cm) ? (modes[new_mode].rtmin >= cm[nb_cm-1].ltmin) : 0))
      new_mode--;
    else
      break;
  }
  if (new_mode < 0) new_mode = -2;

  /* Looking for noise at the beginning of the distribution */
  for (i=1; i < nb_modes; i++)
    if ((modes[i].ltmin - modes[i-1].rtmin) == 1) {
      if (begin_of_noise >= 0 && (i - begin_of_noise) >= 2)
	end_of_noise = i;
      else
	begin_of_noise = i-1;
    }

  /* Looking for max modes */
  for (i = 0; i<nb_modes; i++)
    if (modes[i].pop >= modes[max_mode].pop) max_mode = i;

  /* Looking for the mode in the current distribution corresponding to the
     previous detected mode */
  if (nb_cm) {
    for (i = nb_modes-1 ; i >= 0; i--)
/*       if ((double)modes[i].ltmin + 0.5 <= cm[nb_cm-1].avg &&  */
/* 	  cm[nb_cm-1].avg <= (double)modes[i].rtmin + 0.5) { */
/*       if (modes[i].ltmin <= cm[nb_cm-1].max && cm[nb_cm-1].max <= modes[i].rtmin) { */
/*       if ((double)cm[nb_cm-1].ltmin + 0.5 <= modes[i].avg &&  */
/* 	  modes[i].avg <= (double)cm[nb_cm-1].rtmin + 0.5) { */
      if (cm[nb_cm-1].ltmin <= modes[i].max && modes[i].max <= cm[nb_cm-1].rtmin) {
        last_mode = i;
        break;
      };

    if (!(cm[nb_cm-1].ltmin <= modes[last_mode].max &&
	  modes[last_mode].max <= cm[nb_cm-1].rtmin))
/*       if (!(modes[last_mode].ltmin <= cm[nb_cm-1].max */
/* 	    && cm[nb_cm-1].max <= modes[last_mode].rtmin)) */
/*     if (!((double)modes[last_mode].ltmin + 0.5 <= cm[nb_cm-1].avg && */
/* 	  cm[nb_cm-1].avg <= (double)modes[last_mode].rtmin + 0.5)) */
/*       if (!((double)cm[nb_cm-1].ltmin + 0.5 <= modes[last_mode].avg &&  */
/* 	  modes[last_mode].avg <= (double)cm[nb_cm-1].rtmin + 0.5)) */
      last_mode = new_mode;
  }
  else
    last_mode = new_mode;

/*   printf("new: %d last: %d max: %d EON: %d -> ", new_mode, last_mode, max_mode, end_of_noise); */

  /* If either the new mode or the last detected mode is the max mode, this is
     the capacity mode. It is the best situation, with very little noise and a
     short path. */
  if ((max_mode == new_mode && new_mode >= end_of_noise) || max_mode == last_mode) {
/*     printf("1\n"); */
    cm[nb_cm] = modes[max_mode];
    return nb_cm+1;
  }
  
  /* We try here to detect ICMP rate limiting or filtering. This causes the
     capacity measurements to be non-decreasing. */
  if (max_mode > last_mode && nb_cm) {
/*     printf("2\n"); */
    printf("The previous hop seems to filter or limit the rate of ICMP packets.\n\
The previous result is to be taken with caution.\n");
    cm[nb_cm] = modes[max_mode];
    return nb_cm+1;
  };    

  /* If the new mode counts more (or not too less) measurements than the last
     detected mode and is out of the detected noise, this is the capacity
     mode. */
  if ((1.1*modes[new_mode].pop) >= modes[last_mode].pop && new_mode >= end_of_noise) {
/*     printf("3\n"); */
    cm[nb_cm] = modes[new_mode];
    return nb_cm+1;
  }
  
  /* If the max mode is below the last detected mode and is out of the detected
     noise, this is the capacity mode. */
  if (/* max_mode < new_mode &&  */
      (max_mode > end_of_noise || modes[max_mode].pop >= 0.6*nb_mes) &&
      (modes[max_mode].pop >= 1.25*modes[last_mode].pop)) {
/*       (modes[max_mode].rtmin - modes[max_mode].ltmin <= 3) && */
/*       (modes[max_mode].ltmin - modes[(max_mode-1)%nb_modes].rtmin > 1) && */
/*       (modes[(max_mode+1)%nb_modes].ltmin - modes[max_mode].rtmin > 1)) { */
/*     printf("4\n"); */
    cm[nb_cm] = modes[max_mode];
    return nb_cm+1;
  }

  /* Very last chance: to not disturb measurements, we give as default result
     the last detected mode. */
/*   printf("5\n"); */
  cm[nb_cm] = modes[last_mode];
  return nb_cm+1;
}

/*
int main(int argc, char* argv[]) {
  FILE* fic;
  double* mesures;
  struct mode modes[NB_MODES], capa_modes[NB_MODES], tmp_md;
  int nb_capa_modes = 0;
  int nb_mes, i, j, hop, curr_hop;
  double pas = 0.5, tmp_mes;

  if (argc < 2) {
    printf("Pas de nombre de mesures en argument.\n");
    return 1 ;
  }
  
  nb_mes = atoi(argv[1]);
 
  if (argc == 2) {
    printf("Fichier : stdin\n");
    fic = stdin;
  }
  else {
    printf("Fichier : %s\n", argv[2]);
    if ((fic = fopen(argv[2], "r")) == NULL) {
      printf("Fichier introuvable.\n");
      return 1;
    }
  }
  printf("Nombre de mesures : %d\n", nb_mes);
  
  mesures = calloc(nb_mes, sizeof(double));
//   for (curr_hop = 1; curr_hop <= 6; curr_hop++) {
//     printf("Saut #%d :\n", curr_hop);

    i = 0;
    while (!feof(fic) && i<nb_mes) {
      fscanf(fic, "%d %lf\n", &hop, &tmp_mes);
      // if (hop == curr_hop)
      mesures[i++] = tmp_mes;
    }
    rewind(fic);
    printf("\tNombre de mesures lues : %d\n", i);
    nb_mes = i;

    j = detect_modes(mesures, nb_mes, pas, modes);
    printf("\tNombre de modes : %d\n", j);
    for (i = 0; i<j; i++)
      printf("\t%2d: %7.2f %7.2f %7.2f -> %3d (%.2f)\n",
             i, (modes[i].ltmin+0.5)*pas, (modes[i].max+0.5)*pas, (modes[i].rtmin+0.5)*pas,
             modes[i].pop, modes[i].avg*pas);

    nb_capa_modes = extract_capa_mode(capa_modes, nb_capa_modes, modes, j, nb_mes);
    tmp_md = capa_modes[nb_capa_modes-1];
//     printf("\t ==> CM : %7.2f %7.2f %7.2f -> %3d (%.2f)\n",
//            (tmp_md.ltmin+0.5)*pas, (tmp_md.max+0.5)*pas, (tmp_md.rtmin+0.5)*pas,
//            tmp_md.pop, tmp_md.avg*pas);
//   }

  return 0;
}
*/
